@@ -27,7 +27,7 @@ use Config::IniFiles;
 #inherits Lemonldap::NG::Common::Conf::Backends::SOAP
 #inherits Lemonldap::NG::Common::Conf::Backends::LDAP
 
-our $VERSION = '2.0.15';
+our $VERSION = '2.19.0';
 our $msg     = '';
 our $iniObj;
 
@@ -86,7 +86,7 @@ sub new {
             $msg .= "Error: configStorage: type is not well formed.\n";
         }
         $self->{type} = "Lemonldap::NG::Common::Conf::Backends::$self->{type}"
-          unless $self->{type} =~ /^Lemonldap::/;
+          unless $self->{type} =~ /::/;
         eval "require $self->{type}";
         if ($@) {
             $msg .= "Error: failed to load $self->{type}: \n $@";
@@ -143,9 +143,17 @@ sub saveConf {
     }
 
     $msg .= "Configuration $conf->{cfgNum} stored.\n";
-    if ( $self->{refLocalStorage} ) {
-        $self->setDefault($conf);
-        $self->compactConf($conf);
+    $self->setDefault($conf);
+    $self->setValuesFromEnv($conf);
+    $self->compactConf($conf);
+
+    if (    Lemonldap::NG::Handler::Main->can('tsv')
+        and Lemonldap::NG::Handler::Main->tsv->{msgBrokerWriter} )
+    {
+        Lemonldap::NG::Handler::Main->tsv->{msgBrokerWriter}
+          ->publish( $conf->{eventQueueName}, { action => 'newConf' } );
+    }
+    else {
         eval { Lemonldap::NG::Handler::Main->reload() };
     }
 
@@ -164,10 +172,24 @@ sub saveConf {
 sub getConf {
     my ( $self, $args ) = @_;
     my $res;
+    my $local    = $args->{local};
+    my $cfgNum   = $args->{cfgNum};
+    my $localPrm = $args->{localPrm};
+    my $noCache  = $args->{noCache};
+    my $raw      = $args->{raw};
+    my $allow_cache_for_root =
+      (   $self->{localStorageOptions}
+        ? $self->{localStorageOptions}->{allow_cache_for_root}
+        : 0 );
 
-    # Use only cache to get conf if $args->{local} is set
-    if (    $>
-        and $args->{local}
+    # If running as UID=0, disabled the cache unless explicitely allowed
+    if ( $> == 0 and !$allow_cache_for_root ) {
+        $noCache = 1;
+    }
+
+    # Use only cache to get conf if $local is set
+    if (   !$noCache
+        and $local
         and ref( $self->{refLocalStorage} )
         and $res = $self->{refLocalStorage}->get('conf') )
     {
@@ -177,34 +199,34 @@ sub getConf {
     # Check cfgNum in conf backend
     # Get conf in backend only if a newer configuration is available
     else {
-        $args->{cfgNum} ||= $self->lastCfg;
-        unless ( $args->{cfgNum} ) {
+        $cfgNum ||= $self->lastCfg;
+        unless ($cfgNum) {
             $msg .= "Error: No configuration available in backend.\n";
         }
         my $r;
         unless ( ref( $self->{refLocalStorage} ) ) {
             $msg .= "Get remote configuration (localStorage unavailable).\n";
-            $r = $self->getDBConf($args);
+            $r = $self->getDBConf( { cfgNum => $cfgNum } );
             return undef unless ( $r->{cfgNum} );
-            $self->setDefault( $r, $args->{localPrm} );
+            $self->setDefault( $r, $localPrm );
+            $self->setValuesFromEnv($r);
             $self->compactConf($r);
         }
         else {
             eval { $r = $self->{refLocalStorage}->get('conf') }
-              if ( $> and not $args->{noCache} );
+              if ( !$noCache );
             $msg .= "Warn: $@" if ($@);
 
             if (    ref($r)
                 and $r->{cfgNum}
-                and $args->{cfgNum}
-                and $r->{cfgNum} == $args->{cfgNum} )
+                and $cfgNum
+                and $r->{cfgNum} == $cfgNum )
             {
                 $msg .=
                   "Configuration unchanged, get configuration from cache.\n";
-                $args->{noCache} = 1;
             }
             else {
-                my $r2 = $self->getDBConf($args);
+                my $r2 = $self->getDBConf( { cfgNum => $cfgNum } );
                 unless ( $r2->{cfgNum} ) {
                     $r = $self->{refLocalStorage}->get('conf') unless ($r);
                     $msg .=
@@ -217,13 +239,11 @@ sub getConf {
                     $r = $r2;
                 }
 
-                $self->setDefault( $r, $args->{localPrm} );
-                $self->compactConf($r);
-
-                # Store modified configuration in cache
+                # Store configuration in cache
                 $self->setLocalConf($r)
-                  if ( $self->{refLocalStorage}
-                    and not( $args->{noCache} == 1 or $args->{raw} ) );
+                  if (  $self->{refLocalStorage}
+                    and ( !$noCache or $noCache != 1 )
+                    and !$raw );
             }
         }
 
@@ -231,8 +251,12 @@ sub getConf {
         $res = $r;
     }
 
+    $self->setDefault( $res, $localPrm );
+    $self->setValuesFromEnv($res);
+    $self->compactConf($res);
+
     # Create cipher object and replace variable placeholder
-    unless ( $args->{raw} ) {
+    unless ($raw) {
 
         $self->replacePlaceholders($res) if $self->{useServerEnv};
         eval {
@@ -278,6 +302,13 @@ sub setDefault {
     return $conf;
 }
 
+sub setValuesFromEnv {
+    my ( $self, $conf ) = @_;
+    $conf->{hashedSessionStore} = $ENV{LLNG_HASHED_SESSION_STORE}
+      if $ENV{LLNG_HASHED_SESSION_STORE};
+    return $conf;
+}
+
 ## @method hashRef getLocalConf(string section, string file, int loaddefault)
 # Get configuration from local file
 #
@@ -303,18 +334,21 @@ sub getLocalConf {
         # If default configuration cannot be read
         # - Error if configuration section is requested
         # - Silent exit for other section requests
-        unless ( -r $file ) {
+        if ( open( my $fh, '<', $file ) ) {
+
+            # Parse ini file
+            $cfg = Config::IniFiles->new( -file => $fh, -allowcontinue => 1 );
+            close($fh);
+        }
+        else {
             if ( $section eq CONFSECTION ) {
                 $msg .=
-                  "Cannot read $file to get configuration access parameters.\n";
+"Cannot read $file to get configuration access parameters: $!\n";
                 return $r;
             }
             return $r;
         }
 
-        # Parse ini file
-        $cfg = Config::IniFiles->new( -file => $file, -allowcontinue => 1 );
-
         unless ( defined $cfg ) {
             $msg .= "Local config error: "
               . ( join "\n", @Config::IniFiles::errors ) . "\n";
@@ -339,8 +373,11 @@ sub getLocalConf {
         foreach ( $cfg->Parameters(DEFAULTSECTION) ) {
             $r->{$_} = $cfg->val( DEFAULTSECTION, $_ );
             if ( $_ eq "require" ) {
-                eval { require $r->{$_} };
-                $msg .= "Error: $@" if ($@);
+
+                foreach my $require ( split /[,\s]+/, $r->{$_} ) {
+                    eval { require $require };
+                    $msg .= "Error: $@" if ($@);
+                }
             }
             if (   $r->{$_} =~ /^[{\[].*[}\]]$/
                 || $r->{$_} =~ /^sub\s*{.*}$/ )
@@ -383,7 +420,6 @@ sub getLocalConf {
 # @param $conf Lemonldap::NG configuration hashRef
 sub setLocalConf {
     my ( $self, $conf ) = @_;
-    return unless ($>);
     eval { $self->{refLocalStorage}->set( "conf", $conf ) };
     $msg .= "Warn: $@\n" if ($@);
 }
@@ -391,20 +427,21 @@ sub setLocalConf {
 ## @method hashRef getDBConf(hashRef args)
 # Get configuration from remote storage system.
 # @param $args hashRef that must contains a key "cfgNum" (number of the wanted
-# configuration) and optionaly a key "fields" that points to an array of wanted
-# configuration keys
+# configuration)
 # @return Lemonldap::NG configuration hashRef
 sub getDBConf {
     my ( $self, $args ) = @_;
-    return undef unless $args->{cfgNum};
-    if ( $args->{cfgNum} < 0 ) {
+    my $cfgNum = $args->{cfgNum};
+    return undef unless $cfgNum;
+
+    if ( $cfgNum < 0 ) {
         my @a = $self->available();
-        $args->{cfgNum} =
-            ( @a + $args->{cfgNum} > 0 )
-          ? ( $a[ $#a + $args->{cfgNum} ] )
+        $cfgNum =
+            ( @a + $cfgNum > 0 )
+          ? ( $a[ $#a + $cfgNum ] )
           : $a[0];
     }
-    my $conf = $self->load( $args->{cfgNum} );
+    my $conf = $self->load($cfgNum);
     return undef if $conf == "-1";
     $msg .= "Get configuration $conf->{cfgNum}.\n"
       if ( defined $conf->{cfgNum} );
@@ -510,7 +547,7 @@ sub logError {
 
 sub _substPlaceHolders {
     return $_[0] unless $_[0];
-    $_[0] =~ s/$PlaceHolderRe/$ENV{$1}/geo;
+    $_[0] =~ s/$PlaceHolderRe/$ENV{$1}/ge;
     return $_[0];
 }
 
@@ -521,7 +558,7 @@ sub replacePlaceholders {
     my ( $self, $conf ) = @_;
     if ( ref $conf eq 'HASH' ) {
         foreach my $key ( keys %$conf ) {
-            if ( $key =~ /$PlaceHolderRe/o ) {
+            if ( $key =~ /$PlaceHolderRe/ ) {
                 my $val = $conf->{$key};
                 delete $conf->{$key};
                 my $nk = _substPlaceHolders($key);
@@ -531,7 +568,7 @@ sub replacePlaceholders {
             if ( ref $conf->{$key} ) {
                 $self->replacePlaceholders( $conf->{$key} );
             }
-            elsif ( $conf->{$key} =~ /$PlaceHolderRe/o ) {
+            elsif ( $conf->{$key} =~ /$PlaceHolderRe/ ) {
                 $conf->{$key} = _substPlaceHolders( $conf->{$key} );
             }
         }
@@ -541,7 +578,7 @@ sub replacePlaceholders {
             if ( ref $conf->[$i] ) {
                 $self->replacePlaceholders( $conf->[$i] );
             }
-            elsif ( $conf->[$i] =~ /$PlaceHolderRe/o ) {
+            elsif ( $conf->[$i] =~ /$PlaceHolderRe/ ) {
                 $conf->[$i] = _substPlaceHolders( $conf->[$i] );
             }
         }
@@ -620,9 +657,6 @@ a hash reference as first argument containing 2 optional parameters:
 =item * C<cfgNum => $number>: the number of the configuration wanted. If this
 argument is omitted, the last configuration is returned.
 
-=item * C<fields => [array of names]: the desired fields asked. By default,
-getConf returns all (C<select * from lmConfig>).
-
 =back
 
 =item * B<saveConf>: stores the Lemonldap::NG configuration passed in argument
