@@ -6,7 +6,7 @@ use Mouse;
 use Lemonldap::NG::Common::Conf::Constants;
 use Lemonldap::NG::Common::Conf::ReConstants;
 
-our $VERSION = '2.0.7';
+our $VERSION = '2.19.0';
 
 extends 'Lemonldap::NG::Common::Conf::AccessLib';
 
@@ -198,8 +198,8 @@ sub virtualHosts {
             # If rule contains a comment or an AuthLevel, split them
             if ( $query eq 'locationRules' ) {
                 $res->{comment} = '';
-                $res->{level} = '';
-                $res->{level} = $1 if ( $r =~ s/\(\?#AuthnLevel=(-?\d+)\)// );
+                $res->{level}   = '';
+                $res->{level}   = $1 if ( $r =~ s/\(\?#AuthnLevel=(-?\d+)\)// );
                 if ( $r =~ s/\(\?#(.*?)\)// ) {
                     $res->{title} = $res->{comment} = $1;
                 }
@@ -269,9 +269,7 @@ sub _samlMetaDataNodes {
     my ( $id, $resp ) = ( 1, [] );
 
     # Return all exported attributes if asked
-    if ( $query =~
-        /^saml${type}MetaDataExportedAttributes|samlSPMetaDataMacros$/ )
-    {
+    if ( $query =~ /^saml${type}MetaDataExportedAttributes$/ ) {
         my $pk =
           eval { $self->getConfKey( $req, $query )->{$partner} } // {};
         return $self->sendError( $req, undef, 400 ) if ( $req->error );
@@ -286,6 +284,21 @@ sub _samlMetaDataNodes {
         }
         return $self->sendJSONresponse( $req, $resp );
     }
+    elsif ( $query eq "samlSPMetaDataMacros" ) {
+        my $pk =
+          eval { $self->getConfKey( $req, $query )->{$partner} } // {};
+        return $self->sendError( $req, undef, 400 ) if ( $req->error );
+        foreach my $h ( sort keys %$pk ) {
+            push @$resp,
+              {
+                id    => "saml${type}MetaDataNodes/$partner/$query/" . $id++,
+                title => $h,
+                data  => $pk->{$h},
+                type  => 'keyText',
+              };
+        }
+        return $self->sendJSONresponse( $req, $resp );
+    }
 
     # Simple root keys
     elsif ( $query =~ /^saml${type}MetaDataXML$/ ) {
@@ -380,9 +393,34 @@ sub _oidcMetaDataNodes {
 
     my ( $id, $resp ) = ( 1, [] );
 
+    # Handle RP Attributes
+    if ( $query eq "oidcRPMetaDataExportedVars" ) {
+        my $pk = eval { $self->getConfKey( $req, $query )->{$partner} } // {};
+        return $self->sendError( $req, undef, 400 ) if ( $req->error );
+        foreach my $h ( sort keys %$pk ) {
+
+            # Set default values for type and array
+            my $data = [ split /;/, $pk->{$h} ];
+            unless ( $data->[1] ) {
+                $data->[1] = "string";
+            }
+            unless ( $data->[2] ) {
+                $data->[2] = "auto";
+            }
+            push @$resp,
+              {
+                id    => "oidc${type}MetaDataNodes/$partner/$query/" . $id++,
+                title => $h,
+                data  => $data,
+                type  => 'oidcAttribute',
+              };
+        }
+        return $self->sendJSONresponse( $req, $resp );
+    }
+
     # Return all exported attributes if asked
-    if ( $query =~
-/^(?:oidc${type}MetaDataExportedVars|oidcRPMetaDataOptionsExtraClaims|oidcRPMetaDataMacros)$/
+    elsif ( $query =~
+/^(?:oidc${type}MetaDataExportedVars|oidcRPMetaDataOptionsExtraClaims|oidcRPMetaDataMacros|oidcRPMetaDataScopeRules)$/
       )
     {
         my $pk = eval { $self->getConfKey( $req, $query )->{$partner} } // {};
@@ -544,7 +582,7 @@ sub authChoiceModules {
     unless ($key) {
         my @res;
         foreach my $k ( sort keys %$value ) {
-            my $data = [ split /;/, $value->{$k} ];
+            my $data = [ split /;\s*/, $value->{$k} ];
             if ( $data->[5] ) {
                 my $over;
                 eval { $over = from_json( $data->[5] ) };
@@ -567,7 +605,7 @@ sub authChoiceModules {
         return $self->sendJSONresponse( $req, \@res );
     }
     else {
-        my $r = $value->{$key} ? [ split( /[;\|]/, $value->{$key} ) ] : [];
+        my $r = $value->{$key} ? [ split( /;\s*/, $value->{$key} ) ] : [];
         return $self->sendJSONresponse( $req, { value => $r } );
     }
 }
@@ -697,9 +735,9 @@ sub combModules {
     my $res = [];
     foreach my $mod ( keys %$val ) {
         my $tmp;
-        $tmp->{title} = $mod;
-        $tmp->{id}    = "combModules/$mod";
-        $tmp->{type}  = 'cmbModule';
+        $tmp->{title}      = $mod;
+        $tmp->{id}         = "combModules/$mod";
+        $tmp->{type}       = 'cmbModule';
         $tmp->{data}->{$_} = $val->{$mod}->{$_} foreach (qw(type for));
         my $over = $val->{$mod}->{over} // {};
         $tmp->{data}->{over} = [ map { [ $_, $over->{$_} ] } keys %$over ];
@@ -720,7 +758,8 @@ sub sfExtra {
         $tmp->{id}         = "sfExtra/$mod";
         $tmp->{type}       = 'sfExtra';
         $tmp->{data}->{$_} = $val->{$mod}->{$_}
-          foreach (qw(type rule logo label));
+          foreach (qw(type rule regrule logo level label));
+        $tmp->{data}->{register} = $val->{$mod}->{register} ? \1 : \0;
         my $over = $val->{$mod}->{over} // {};
         $tmp->{data}->{over} = [ map { [ $_, $over->{$_} ] } keys %$over ];
         push @$res, $tmp;
@@ -744,24 +783,90 @@ sub metadata {
     if ( $req->params('full') and $req->params('full') !~ NO ) {
         my $c = $self->getConfKey( $req, 'cfgNum' );
         return $self->sendError( $req, undef, 400 ) if ( $req->error );
-        if ( $self->can('userId') ) {
-            $self->userLogger->notice( 'User '
-                  . $self->userId($req)
-                  . ' ask for full configuration '
-                  . $c );
+        if ( my $userid = $self->_userId($req) ) {
+            $self->auditLog(
+                $req,
+                message => (
+                    'User ' . $userid . ' ask for full configuration ' . $c
+                ),
+                code   => "CONFIG_DOWNLOADED",
+                user   => $userid,
+                cfgNum => $c
+            );
         }
         else {
-            $self->logger->info("REST request to get full configuration $c");
+            $self->auditLog(
+                $req,
+                message => "REST request to get full configuration $c",
+                code    => "CONFIG_DOWNLOADED",
+                cfgNum  => $c
+            );
         }
         return $self->sendJSONresponse(
             $req,
             $self->currentConf,
-            forceJSON => 1,
-            headers   => [
+            pretty  => 1,
+            headers => [
                 'Content-Disposition' => "Attachment; filename=lmConf-$c.json"
             ],
         );
     }
+    elsif ( my $oidc = $req->params('oidcMetadata')
+        or $req->params('samlMetadata') )
+    {
+        my $c    = $self->getConfKey( $req, 'cfgNum' );
+        my $type = $oidc ? 'OIDC' : 'SAML';
+        return $self->sendError( $req, undef, 400 ) if ( $req->error );
+        if ( $self->can('userId') ) {
+            $self->userLogger->notice( 'User '
+                  . $self->userId($req)
+                  . " ask for $type metadata "
+                  . $c );
+        }
+        else {
+            $self->logger->info("REST request to get $type metadata $c");
+        }
+        if ($oidc) {
+            require Lemonldap::NG::Common::OpenIDConnect::Metadata;
+            my $path = $self->currentConf->{issuerDBOpenIDConnectPath};
+            $path =~ s#^.*?(\w+).*?$#$1/#;
+            return $self->sendJSONresponse(
+                $req,
+                Lemonldap::NG::Common::OpenIDConnect::Metadata->metadataDoc(
+                    $self->currentConf->{oidcServiceMetaDataIssuer}
+                      || $self->p->portal,
+                    $self->currentConf,
+                    $path,
+                ),
+                pretty  => 1,
+                headers => [
+                    'content-Type'        => 'application/json',
+                    'Content-Disposition' =>
+                      "Attachment; filename=openid-configuration.json"
+                ],
+            );
+        }
+        else {
+            require Lemonldap::NG::Common::Conf::SAML::Metadata;
+            if ( my $metadata =
+                Lemonldap::NG::Common::Conf::SAML::Metadata->new() )
+            {
+                my $s =
+                  $metadata->serviceToXML(
+                    { %{ $self->currentConf }, portal => $self->p->portal },
+                    'all' );
+                return [
+                    200,
+                    [
+                        'Content-Type'   => 'application/xml',
+                        'Content-Length' => length($s),
+                    ],
+                    [$s]
+                ];
+            }
+            return $self->p->sendError( $req, 'Unable to build Metadata', 500 );
+        }
+    }
     else {
         my $res = {};
         $res->{cfgNum} = $self->getConfKey( $req, 'cfgNum' );
@@ -773,23 +878,35 @@ sub metadata {
         }
 
         # Find next and previous conf
-        my @a  = $self->confAcc->available;
-        my $id = -1;
+        my @a     = $self->confAcc->available;
+        my $id    = -1;
         my ($ind) = map { $id++; $_ == $res->{cfgNum} ? ($id) : () } @a;
         if ($ind) { $res->{prev} = $a[ $ind - 1 ]; }
         if ( defined $ind and $ind < $#a ) {
             $res->{next} = $a[ $ind + 1 ];
         }
-        if ( $self->can('userId') ) {
-            $self->userLogger->info( 'User '
-                  . $self->userId($req)
-                  . ' ask for configuration metadata ('
-                  . $res->{cfgNum}
-                  . ')' );
+        if ( my $userid = $self->_userId($req) ) {
+            $self->auditLog(
+                $req,
+                code    => "CONFIG_GET_METADATA",
+                user    => $userid,
+                cfgNum  => $res->{cfgNum},
+                message => (
+                        'User '
+                      . $userid
+                      . ' ask for configuration metadata ('
+                      . $res->{cfgNum} . ')'
+                )
+            );
         }
         else {
-            $self->logger->info(
-                "REST request to get configuration metadata ($res->{cfgNum})");
+            $self->auditLog(
+                $req,
+                code    => "CONFIG_GET_METADATA",
+                cfgNum  => $res->{cfgNum},
+                message =>
+                  "REST request to get configuration metadata ($res->{cfgNum})",
+            );
         }
         return $self->sendJSONresponse( $req, $res );
     }
@@ -810,12 +927,22 @@ sub getKey {
     unless ($key) {
         return $self->metadata($req);
     }
-    if ( $self->can('userId') ) {
-        $self->userLogger->info(
-            'User ' . $self->userId($req) . " asks for key $key" );
+    if ( my $userid = $self->_userId($req) ) {
+        $self->auditLog(
+            $req,
+            code    => "CONFIG_GET_KEY",
+            message => ( 'User ' . $userid . " asks for key $key" ),
+            user    => $userid,
+            key     => $key,
+        );
     }
     else {
-        $self->logger->info("REST request to get configuration key $key");
+        $self->auditLog(
+            $req,
+            code    => "CONFIG_GET_KEY",
+            message => ("REST request to get configuration key $key"),
+            key     => $key,
+        );
     }
     my $value = $self->getConfKey( $req, $key );
     return $self->sendError( $req, undef, 400 ) if ( $req->error );
@@ -867,4 +994,12 @@ sub getKey {
     # TODO authParam key
 }
 
+sub _userId {
+    my ( $self, $req ) = @_;
+    if ( $self->can('p') and $self->p->can('userId') ) {
+        return $self->p->userId($req);
+    }
+    return;
+}
+
 1;
