@@ -11,7 +11,7 @@ use Time::Local;
 use DBI;
 use Encode;
 
-our $VERSION = '2.0.7';
+our $VERSION = '2.21.0';
 
 extends 'Lemonldap::NG::Common::Notifications';
 
@@ -42,21 +42,6 @@ has dbiUser => (
 
 has dbiPassword => ( is => 'ro', default => '' );
 
-# Database handle object
-has _dbh => (
-    is      => 'rw',
-    lazy    => 1,
-    builder => sub {
-        my $self = shift;
-        my $r    = DBI->connect_cached(
-            $self->{dbiChain}, $self->{dbiUser},
-            $self->{dbiPassword}, { RaiseError => 1 }
-        );
-        $self->logger->error($DBI::errstr) unless ($r);
-        return $r;
-    }
-);
-
 # Current query
 has sth => ( is => 'rw' );
 
@@ -92,6 +77,36 @@ sub get {
     return $result;
 }
 
+# Returns accepted notifications corresponding to the user $uid.
+# If $ref is set, returns only notification corresponding to this reference.
+sub getAccepted {
+    my ( $self, $uid, $ref ) = @_;
+    return () unless ($uid);
+    $self->_execute(
+        "SELECT * FROM "
+          . $self->dbiTable
+          . " WHERE uid=? AND ref=? ORDER BY date",
+        $uid,
+        ( $ref ? $ref : () )
+    ) or return ();
+    my $result;
+    while ( my $h = $self->sth->fetchrow_hashref() ) {
+
+        # Get XML message
+        my $xml = $h->{xml};
+
+        # Decode it to get the correct uncoded string
+        Encode::from_to( $xml, "utf8", "iso-8859-1", Encode::FB_CROAK );
+
+        # Store message in result
+        my $identifier =
+          &getIdentifier( $self, $h->{uid}, $h->{ref}, $h->{date} );
+        $result->{$identifier} = $xml;
+    }
+    $self->logger->warn( $self->sth->err() ) if ( $self->sth->err() );
+    return $result;
+}
+
 ## @method hashref getAll()
 # Return all pending notifications.
 # @return hashref where keys are internal reference and values are hashref with
@@ -107,7 +122,7 @@ sub getAll {
             date      => $h->{date},
             uid       => $h->{uid},
             ref       => $h->{ref},
-            condition => $h->{condition}
+            condition => $h->{cond}
         };
     }
     $self->logger->warn( $self->sth->err() ) if ( $self->sth->err() );
@@ -120,16 +135,14 @@ sub getAll {
 # keys date, uid, ref and condition.
 sub getExisting {
     my $self = shift;
-    $self->_execute( 'SELECT * FROM '
-          . $self->dbiTable
-          . ' ORDER BY date' );
+    $self->_execute( 'SELECT * FROM ' . $self->dbiTable . ' ORDER BY date' );
     my $result;
     while ( my $h = $self->sth->fetchrow_hashref() ) {
         $result->{"$h->{date}#$h->{uid}#$h->{ref}"} = {
             date      => $h->{date},
             uid       => $h->{uid},
             ref       => $h->{ref},
-            condition => $h->{condition}
+            condition => $h->{cond}
         };
     }
     $self->logger->warn( $self->sth->err() ) if ( $self->sth->err() );
@@ -194,7 +207,12 @@ sub purge {
 # @return true if succeed
 sub newNotif {
     my ( $self, $date, $uid, $ref, $condition, $xml ) = @_;
-
+    my @t = split( /\D+/, $date );
+    $t[1]--;
+    eval {
+        timelocal( $t[5] || 0, $t[4] || 0, $t[3] || 0, $t[2], $t[1], $t[0] );
+    };
+    return ( 0, "Bad date" ) if ($@);
     my $res =
       $condition =~ /.+/
       ? $self->_execute( 'INSERT INTO '
@@ -220,10 +238,21 @@ sub getDone {
     my $result;
     while ( my $h = $self->sth->fetchrow_hashref() ) {
         my @t = split( /\D+/, $h->{date} );
-        my $done =
-          timelocal( $t[5] || 0, $t[4] || 0, $t[3] || 0, $t[2], $t[1], $t[0] );
-        $result->{"$h->{date}#$h->{uid}#$h->{ref}"} =
-          { notified => $done, uid => $h->{uid}, ref => $h->{ref}, };
+        $t[1]--;
+        my $done = eval {
+            timelocal( $t[5] || 0, $t[4] || 0, $t[3] || 0, $t[2], $t[1],
+                $t[0] );
+        };
+        if ($@) {
+            $self->logger->warn("Bad date: $h->{date}");
+            return {};
+        }
+        $result->{"$h->{date}#$h->{uid}#$h->{ref}"} = {
+            notified  => $done,
+            uid       => $h->{uid},
+            ref       => $h->{ref},
+            condition => $h->{cond}
+        };
     }
     $self->logger->warn( $self->sth->err() ) if ( $self->sth->err() );
     return $result;
@@ -235,6 +264,7 @@ sub getDone {
 sub _execute {
     my ( $self, $query, @args ) = @_;
     my $dbh = $self->_dbh or die "DB connection unavailable";
+    $self->logger->debug("Notification DBI proccessing query: $query");
     unless ( $self->sth( $dbh->prepare($query) ) ) {
         $self->logger->warn( $dbh->errstr() );
         return 0;
@@ -258,5 +288,27 @@ sub getIdentifier {
     return $date . "#" . $uid . "#" . $ref;
 }
 
-1;
+sub _dbh {
+    my $self = shift;
+    return $self->{_dbh} if ( $self->{_dbh} and $self->{_dbh}->ping );
+    $self->logger->debug("Notification DBI connection lost, getting a new one");
+    $self->{_dbh} = DBI->connect_cached( $self->{dbiChain}, $self->{dbiUser},
+        $self->{dbiPassword}, { RaiseError => 1, AutoCommit => 1, } );
+    if ( $self->{dbiChain} =~ /^dbi:sqlite/i ) {
+        $self->{_dbh}->{sqlite_unicode} = 1;
+    }
+    elsif ( $self->{dbiChain} =~ /^dbi:mysql/i ) {
+        eval {
+            $self->{_dbh}->{mysql_enable_utf8} = 1;
+            $self->{_dbh}->do("set names 'utf8'");
+        };
+    }
 
+    elsif ( $self->{dbiChain} =~ /^dbi:pg/i ) {
+        $self->{_dbh}->{pg_enable_utf8} = 1;
+    }
+
+    return $self->{_dbh};
+}
+
+1;
