@@ -1,8 +1,9 @@
 package Lemonldap::NG::Handler::Lib::OAuth2;
+use Lemonldap::NG::Common::JWT qw(getAccessTokenSessionId);
 
 use strict;
 
-our $VERSION = '2.0.8';
+our $VERSION = '2.19.0';
 
 sub retrieveSession {
     my ( $class, $req, $id ) = @_;
@@ -10,22 +11,24 @@ sub retrieveSession {
 
     # Retrieve regular session if this is not an offline access token
     unless ($offlineId) {
-        my $data = {
-            %{
-                $class->Lemonldap::NG::Handler::Main::retrieveSession( $req,
-                    $id )
-            },
-            $class->_getTokenAttributes($req)
-        };
+        my $data =
+          $class->Lemonldap::NG::Handler::Main::retrieveSession( $req, $id );
+        if ( ref($data) eq "HASH" ) {
+            $data = { %{$data}, $class->_getTokenAttributes($req) };
 
-        # Update cache
-        $class->data($data);
+            # Update cache
+            $class->data($data);
+        }
+        else {
+            $req->data->{oauth2_error} = 'invalid_token';
+        }
         return $data;
     }
 
     # 2. Get the session from cache or backend
     my $session = $req->data->{session} = (
         Lemonldap::NG::Common::Session->new( {
+                hashStore            => $class->tsv->{hashedSessionStore},
                 storageModule        => $class->tsv->{oidcStorageModule},
                 storageModuleOptions => $class->tsv->{oidcStorageOptions},
                 cacheModule          => $class->tsv->{sessionCacheModule},
@@ -63,7 +66,7 @@ sub retrieveSession {
         return $data;
     }
     else {
-        $class->logger->info("Session $offlineId can't be retrieved");
+        $class->logger->info("Session $offlineId can not be retrieved");
         $class->logger->info( $session->error );
 
         return 0;
@@ -77,22 +80,32 @@ sub fetchId {
     my $authorization = $req->{env}->{HTTP_AUTHORIZATION};
 
     if ( $authorization
-        and ( ($access_token) = ( $authorization =~ /^Bearer (.+)$/i ) ) )
+        and ( $access_token = ( $authorization =~ /^Bearer (.+)$/i )[0] ) )
     {
-        $class->logger->debug( 'Found OAuth2 access token ' . $access_token );
+        $class->logger->debug("Found OAuth2 Access Token ($access_token)");
     }
     else {
         return $class->Lemonldap::NG::Handler::Main::fetchId($req);
     }
 
     # Get access token session
-    my $infos = $class->getOIDCInfos($access_token);
+    my $access_token_sid = getAccessTokenSessionId($access_token);
+    unless ($access_token_sid) {
+        $req->data->{oauth2_error} = 'invalid_token';
+        return;
+    }
+    my $infos = $class->getOIDCInfos($access_token_sid);
+    unless ($infos) {
+        $req->data->{oauth2_error} = 'invalid_token';
+        return;
+    }
 
     # Store scope and rpid for future session attributes
     if ( $infos->{rp} ) {
         my $rp = $infos->{rp};
-        $req->data->{_scope}         = $infos->{scope};
-        $req->data->{_clientConfKey} = $rp;
+        $req->data->{_scope}           = $infos->{scope};
+        $req->data->{_oidc_grant_type} = $infos->{grant_type};
+        $req->data->{_clientConfKey}   = $rp;
         if (    $class->tsv->{oauth2Options}->{$rp}
             and $class->tsv->{oauth2Options}->{$rp}->{clientId} )
         {
@@ -129,6 +142,7 @@ sub getOIDCInfos {
 
     # Get the session
     my $oidcSession = Lemonldap::NG::Common::Session->new( {
+            hashStore            => $class->tsv->{hashedSessionStore},
             storageModule        => $class->tsv->{oidcStorageModule},
             storageModuleOptions => $class->tsv->{oidcStorageOptions},
             cacheModule          => $class->tsv->{sessionCacheModule},
@@ -141,10 +155,24 @@ sub getOIDCInfos {
     unless ( $oidcSession->error ) {
         $class->logger->debug("Get OIDC session $id");
 
+        # Verify that session is valid
+        unless ( $oidcSession->data->{_utime} ) {
+            $class->logger->error("_utime missing from Access Token session");
+            return;
+        }
+
+        my $ttl = $class->tsv->{timeout} - time + $oidcSession->data->{_utime};
+        $class->logger->debug("Session TTL = $ttl");
+
+        if ( time - $oidcSession->data->{_utime} > $class->tsv->{timeout} ) {
+            $class->logger->info("Access Token session $id expired");
+            return;
+        }
+
         $infos = { %{ $oidcSession->data } };
     }
     else {
-        $class->logger->info("OIDC Session $id can't be retrieved");
+        $class->logger->info("OIDC session $id can not be retrieved");
         $class->logger->info( $oidcSession->error );
     }
 
@@ -168,7 +196,7 @@ sub goToPortal {
 sub _getTokenAttributes {
     my ( $class, $req ) = @_;
     my %res;
-    for my $attr (qw/_scope _clientConfKey _clientId/) {
+    for my $attr (qw/_scope _clientConfKey _clientId _oidc_grant_type/) {
         if ( $req->data->{$attr} ) {
             $res{$attr} = $req->data->{$attr};
         }
