@@ -25,6 +25,7 @@ has firstLmConfFile            => ( isa => 'Str', is => 'ro', required => 1 );
 has reverseTreeFile            => ( isa => 'Str', is => 'ro', required => 1 );
 has portalConstantsFile        => ( isa => 'Str', is => 'ro', required => 1 );
 has handlerStatusConstantsFile => ( isa => 'Str', is => 'ro', required => 1 );
+has docConstantsFile           => ( isa => 'Str', is => 'ro', required => 1 );
 
 my @managerAttrKeys = qw(keyTest keyMsgFail select type test msgFail default);
 my $format          = 'Creating %-69s: ';
@@ -188,7 +189,7 @@ our \$specialNodeHash = {
     samlIDPMetaDataNodes => [qw(samlIDPMetaDataXML samlIDPMetaDataExportedAttributes samlIDPMetaDataOptions)],
     samlSPMetaDataNodes  => [qw(samlSPMetaDataXML samlSPMetaDataExportedAttributes samlSPMetaDataOptions samlSPMetaDataMacros)],
     oidcOPMetaDataNodes  => [qw(oidcOPMetaDataJSON oidcOPMetaDataJWKS oidcOPMetaDataOptions oidcOPMetaDataExportedVars)],
-    oidcRPMetaDataNodes  => [qw(oidcRPMetaDataOptions oidcRPMetaDataExportedVars oidcRPMetaDataOptionsExtraClaims oidcRPMetaDataMacros)],
+    oidcRPMetaDataNodes  => [qw(oidcRPMetaDataOptions oidcRPMetaDataExportedVars oidcRPMetaDataOptionsExtraClaims oidcRPMetaDataMacros oidcRPMetaDataScopeRules)],
     casSrvMetaDataNodes  => [qw(casSrvMetaDataOptions casSrvMetaDataExportedVars)],
     casAppMetaDataNodes  => [qw(casAppMetaDataOptions casAppMetaDataExportedVars casAppMetaDataMacros)],
 };
@@ -207,7 +208,8 @@ EOF
     foreach (@simpleHashKeys) {
         $ra->add($_);
     }
-    print F "our \$simpleHashKeys = '" . $ra->as_string . "';\n"
+    print F "our \$simpleHashKeys = '"
+      . $ra->as_string . "';\n"
       . "our \$specialNodeKeys = '${ignoreKeys}s';\n";
     foreach ( sort keys %cnodesRe ) {
         print F "our \$${_}Keys = '$cnodesRe{$_}';\n";
@@ -236,9 +238,8 @@ EOF
     printf STDERR $format, $self->defaultValuesFile;
     $defaultValues->{locationRules} = $attributes->{locationRules}->{default};
     foreach ( keys %$attributes ) {
-        if (    not /(?:MetaData|vhost)/
-            and $attributes->{$_}->{default}
-            and not $reverseScanResult->{$_} )
+        if ( not /(?:MetaData|vhost)/
+            and $attributes->{$_}->{default} )
         {
             $defaultValues->{$_} = $attributes->{$_}->{default};
         }
@@ -467,7 +468,7 @@ sub buildPortalConstants() {
 
     printf STDERR $format, $self->portalConstantsFile;
     open( F, '>', $self->portalConstantsFile ) or die($!);
-    my $urire = $RE{URI}{HTTP}{ -scheme=>qr/https?/ }{-keep};
+    my $urire = $RE{URI}{HTTP}{ -scheme => qr/https?/ }{-keep};
     $urire =~ s/([\$\@])/\\$1/g;
     my $content = <<EOF;
 # This file is generated by $module. Don't modify it by hand
@@ -557,6 +558,54 @@ EOF
     print F $dst;
     close F;
     print STDERR "done\n";
+
+    printf STDERR $format, $self->docConstantsFile;
+
+    # Doc error code list
+    my %details;
+    if ( open F, 'lemonldap-ng-portal/site/htdocs/static/languages/en.json' ) {
+        local $/ = undef, my $tmp = JSON::from_json(<F>);
+        close F;
+        %details = map { s/^\s+//; $_ } %$tmp;
+    }
+    $content = <<EOF;
+..
+   This file is generated by $module. Don't modify it by hand
+
+Error codes list
+================
+
+.. note::
+
+    This page references all Portal error codes.
+
+.. csv-table::
+   :header: "Error label", "Error number", "Detail"
+   :delim: ,
+   :widths: auto
+
+
+EOF
+
+    for my $key (
+        sort { $portalConstants{$a} <=> $portalConstants{$b} }
+        keys %portalConstants
+      )
+    {
+
+        $content .=
+            "    ``"
+          . $key . "``,"
+          . $portalConstants{$key} . ','
+          . ( %details ? '"' . $details{"PE$portalConstants{$key}"} . '"' : '' )
+          . "\n";
+    }
+
+    open( F, '>', $self->docConstantsFile ) or die($!);
+    print F $content;
+    close F;
+
+    print STDERR "done\n";
 }
 
 sub mydump {
@@ -596,7 +645,7 @@ sub scanTree {
         # Subnode
         elsif ( ref($leaf) ) {
             $jleaf->{title} = $jleaf->{id} = $leaf->{title};
-            $jleaf->{type} = $leaf->{form} if ( $leaf->{form} );
+            $jleaf->{type}  = $leaf->{form} if ( $leaf->{form} );
             if ( $leaf->{title} =~ /^((?:oidc|saml|cas)Service)MetaData$/ ) {
                 no strict 'refs';
                 my @tmp = $self->scanLeaf( $leaf->{nodes} );
@@ -650,8 +699,9 @@ sub scanTree {
             if ( $leaf =~ s/^\*// ) {
                 push @angularScopeVars, [ $leaf, "$path._nodes[$ord]" ];
             }
-            push @sessionTypes, $1
-              if ( $leaf =~ /^(.*)(?<!notification)StorageOptions$/ );
+            if ( $leaf =~ /^(.*)(?<!notification)StorageOptions$/ ) {
+                push @sessionTypes, $1;
+            }
             my $attr = $attributes->{$leaf} or die("Missing attribute $leaf");
 
             #print STDERR "| $attr->{documentation}  |  $leaf  |\n";
@@ -677,6 +727,7 @@ sub scanTree {
                     my $type = $attr->{type};
                     $type =~ s/Container//;
                     foreach my $k ( sort keys( %{ $attr->{default} } ) ) {
+
                         # Special handling for oidcAttribute
                         my $default = $attr->{default}->{$k};
                         if ( $attr->{type} eq 'oidcAttributeContainer' ) {
@@ -703,9 +754,9 @@ sub scanTree {
                 push @cnodesKeys, $leaf;
             }
 
-            # issue 2439
-            # FIXME: in future versions, oidcOPMetaDataJSON and samlIDPMetaDataXML shoud
-            # behave the same
+   # issue 2439
+   # FIXME: in future versions, oidcOPMetaDataJSON and samlIDPMetaDataXML should
+   # behave the same
             if ( $leaf =~ /^oidcOPMetaData(?:JSON|JWKS)$/ ) {
                 push @simpleHashKeys, $leaf;
             }
