@@ -5,6 +5,7 @@ use Crypt::URandom;
 use Mouse;
 use Data::Dumper;
 use JSON;
+use Hash::Merge::Simple;
 use Lemonldap::NG::Common::Conf::ReConstants;
 
 our $VERSION = '2.0.12';
@@ -24,8 +25,8 @@ has cfgNum => (
 
 has log    => ( is => 'rw' );
 has req    => ( is => 'ro' );
-has sep    => ( is => 'rw', isa => 'Str', default => '/' );
-has format => ( is => 'rw', isa => 'Str', default => "%-25s | %-25s | %-25s" );
+has sep    => ( is => 'rw', isa => 'Str',  default => '/' );
+has format => ( is => 'rw', isa => 'Str',  default => "%-25s | %-25s | %-25s" );
 has yes    => ( is => 'rw', isa => 'Bool', default => 0 );
 has safe   => ( is => 'rw', isa => 'Bool', default => 0 );
 has force  => ( is => 'rw', isa => 'Bool', default => 0 );
@@ -138,6 +139,73 @@ sub del {
     return $self->_save($new);
 }
 
+sub merge {
+    my ( $self, @files ) = @_;
+    die 'merge requires at least one file' unless (@files);
+    require Clone;
+    my $old = Clone::clone( $self->mgr->hLoadedPlugins->{conf}->currentConf );
+    my $new;
+    my @valid_files;
+
+    foreach my $file (@files) {
+
+        my $merging;
+        if ( $file =~ /\.ya?ml/ ) {
+            eval {
+                require YAML;
+                $merging = YAML::LoadFile($file);
+            };
+            if ($@) {
+                print STDERR "Skipping invalid YAML file $file: " . $@;
+            }
+        }
+        else {
+            eval {
+                local $/ = undef;
+                open my $fh, '<', $file or die $!;
+                $merging = from_json(<$fh>);
+            };
+            if ($@) {
+                print STDERR "Skipping invalid JSON file $file: " . $@;
+            }
+        }
+        if ( ref($merging) eq "HASH" ) {
+            $new = Hash::Merge::Simple::merge( $new, $merging );
+            push @valid_files, $file;
+        }
+    }
+    die "Nothing to do" unless ref($new) eq "HASH";
+
+    unless ( $self->yes ) {
+        print "Merge configuration changes from "
+          . join( " ", @valid_files ) . " \n";
+        print Dumper($new);
+        print "Confirm (N/y)? ";
+        my $c = <STDIN>;
+        unless ( $c =~ /^y(?:es)?$/ ) {
+            die "Aborting";
+        }
+    }
+
+    $new = Hash::Merge::Simple::merge( $old, $new );
+    _clean_hash_undef($new);
+    return $self->_save($new);
+}
+
+# Remove key => undef from hashes
+# This allows you to remove keys from the config
+sub _clean_hash_undef {
+    my ($hash) = @_;
+    for my $key ( keys %$hash ) {
+        if ( !defined( $hash->{$key} ) ) {
+            delete $hash->{$key};
+        }
+        if ( ref( $hash->{$key} ) eq "HASH" ) {
+            _clean_hash_undef( $hash->{$key} );
+        }
+    }
+}
+
 sub addKey {
     my $self = shift;
     unless ( @_ % 3 == 0 ) {
@@ -149,7 +217,7 @@ sub addKey {
         my $root   = shift;
         my $newKey = shift;
         my $value  = shift;
-        unless ( $root =~ /$simpleHashKeys$/o or $root =~ /$sep/o ) {
+        unless ( $root =~ /$simpleHashKeys$/ or $root =~ /$sep/ ) {
             die "$root is not a simple hash. Aborting";
         }
         $self->logger->info("CLI: Append key $root/$newKey $value");
@@ -190,7 +258,7 @@ sub delKey {
     while (@_) {
         my $root = shift;
         my $key  = shift;
-        unless ( $root =~ /$simpleHashKeys$/o or $root =~ /$sep/o ) {
+        unless ( $root =~ /$simpleHashKeys$/ or $root =~ /$sep/ ) {
             die "$root is not a simple hash. Aborting";
         }
         $self->logger->info("CLI: Remove key $root/$key");
@@ -317,8 +385,8 @@ sub lastCfg {
 
 sub save {
     my ($self) = @_;
-    my $conf = $self->jsonResponse( '/confs/' . $self->cfgNum, 'full=1' );
-    my $json = JSON->new->indent->canonical;
+    my $conf   = $self->jsonResponse( '/confs/' . $self->cfgNum, 'full=1' );
+    my $json   = JSON->new->indent->canonical;
     print $json->encode($conf);
 }
 
@@ -333,8 +401,7 @@ sub restore {
         $conf = join '', <STDIN>;
     }
     else {
-        die "Unable to read $file" unless ( -r $file );
-        open( my $f, $file ) or die $!;
+        open( my $f, '<', $file ) or die "Unable to read $file: $!";
         $conf = join '', <$f>;
         close $f;
         die "Empty or malformed file $file" unless ( $conf =~ /\w/s );
@@ -342,8 +409,33 @@ sub restore {
     $self->logger->info("CLI: Restore conf.");
     my $res = $self->_post( '/confs/raw', '', IO::String->new($conf),
         'application/json', length($conf) );
-    use Data::Dumper;
-    print STDERR Dumper($res);
+
+    my $response = eval { from_json( $res->[2]->[0], { allow_nonref => 1 } ) };
+    if ( $res->[0] == 200 and $response->{result} == 1 ) {
+        print STDERR
+          "Configuration has been restored as version $response->{cfgNum}\n";
+
+    }
+    else {
+        if ( my $message = $response->{error} ) {
+            print STDERR "Failed to restore configuration: $message\n";
+        }
+        else {
+            print STDERR "The supplied configuration could not be restored"
+              . " because it contains errors:\n";
+
+            my @errors = eval {
+                map { $_->{message} } @{ $response->{details}->{__errors__} };
+            };
+            if (@errors) {
+                print STDERR " - $_\n" for @errors;
+            }
+            else {
+                use Data::Dumper;
+                print Dumper($res);
+            }
+        }
+    }
 }
 
 sub rollback {
@@ -404,9 +496,9 @@ sub _getKey {
 
 sub _setKey {
     my ( $self, $conf, $key, $value ) = @_;
-    my $sep = $self->sep;
+    my $sep    = $self->sep;
     my (@path) = split $sep, $key;
-    my $last = pop @path;
+    my $last   = pop @path;
     while ( my $next = shift @path ) {
         $conf = $conf->{$next};
     }
@@ -512,11 +604,11 @@ sub run {
     }
     my $action = shift;
     unless ( $action =~
-/^(?:get|set|del|addKey|delKey|addPostVars|delPostVars|save|restore|rollback)$/
+/^(?:get|set|del|addKey|delKey|addPostVars|delPostVars|merge|save|restore|rollback)$/
       )
     {
         die
-"Unknown action $action. Only get, set, del, addKey, delKey, addPostVars, delPostVars, save, restore, rollback allowed";
+"Unknown action $action. Only get, set, del, addKey, delKey, addPostVars, delPostVars, merge, save, restore, rollback allowed";
     }
 
     unless ( $action eq "restore" ) {
