@@ -2,10 +2,12 @@ package Lemonldap::NG::Manager::Conf::Zero;
 
 use strict;
 
-our $VERSION = '2.0.9';
+our $VERSION = '2.18.0';
 
 sub zeroConf {
-    my ( $domain, $sessionDir, $persistentSessionDir, $notificationDir, $cacheDir ) = @_;
+    my ( $domain, $sessionDir, $persistentSessionDir, $notificationDir,
+        $cacheDir )
+      = @_;
     $domain               ||= 'example.com';
     $sessionDir           ||= '/var/lib/lemonldap-ng/sessions';
     $persistentSessionDir ||= '/var/lib/lemonldap-ng/psessions';
@@ -34,7 +36,7 @@ sub zeroConf {
                         'uri'         => "http://manager.$domain/sessions.html",
                         'description' => 'Explore WebSSO sessions',
                         'logo'        => 'database.png',
-                        'name'        => 'Sessions explorer'
+                        'name'        => 'Sessions Explorer'
                     }
                 },
                 'catname'       => 'Administration',
@@ -45,7 +47,7 @@ sub zeroConf {
                         'uri' => "http://manager.$domain/notifications.html",
                         'display' => 'auto',
                         'logo'    => 'database.png',
-                        'name'    => 'Notifications explorer'
+                        'name'    => 'Notifications Explorer'
                     }
                 },
                 'type' => 'category'
@@ -66,7 +68,7 @@ sub zeroConf {
                 'officialwebsite' => {
                     'options' => {
                         'description' => 'Official LemonLDAP::NG Website',
-                        'uri'         => 'http://lemonldap-ng.org/',
+                        'uri'         => 'https://lemonldap-ng.org/',
                         'display'     => 'on',
                         'logo'        => 'network.png',
                         'name'        => 'Official Website'
@@ -104,7 +106,7 @@ sub zeroConf {
         },
         'cfgNum'               => 0,
         'globalStorageOptions' => {
-            'Directory' => $sessionDir,
+            'Directory'      => $sessionDir,
             'generateModule' =>
               'Lemonldap::NG::Common::Apache::Session::Generate::SHA256',
             'LockDirectory' => "$sessionDir/lock"
@@ -148,7 +150,6 @@ sub zeroConf {
             }
         },
         'registerDB'          => 'Demo',
-        'registerUrl'         => "http://auth.$domain/register",
         'portal'              => "http://auth.$domain/",
         'notificationStorage' => 'File',
         'locationRules'       => {
@@ -167,7 +168,7 @@ sub zeroConf {
             },
             "manager.$domain" => {
                 'default' => 'inGroup("timelords") or $uid eq "rtyler"',
-'(?#Configuration)^/(.*?\.(fcgi|psgi)/)?(manager\.html|confs/|$)'
+'(?#Configuration)^/(.*?\.(fcgi|psgi)/)?(manager\.html|confs|prx/|$)'
                   => 'inGroup("timelords")',
                 '(?#Sessions)/(.*?\.(fcgi|psgi)/)?sessions' =>
                   'inGroup("timelords") or $uid eq "rtyler"',
@@ -175,16 +176,16 @@ sub zeroConf {
                   'inGroup("timelords") or $uid eq "rtyler"',
             }
         },
-        'whatToTrace'   => '_whatToTrace',
-        'securedCookie' => 0,
-        'cookieName'    => 'lemonldap',
-        'cfgAuthor'     => 'The LemonLDAP::NG team',
-        'cfgVersion'    => $VERSION,
-        'exportedVars'  => {},
-        'portalSkin'    => 'bootstrap',
+        'whatToTrace'          => '_whatToTrace',
+        'securedCookie'        => 0,
+        'cookieName'           => 'lemonldap',
+        'cfgAuthor'            => 'The LemonLDAP::NG team',
+        'cfgDate'              => '1627287638',
+        'cfgVersion'           => $VERSION,
+        'exportedVars'         => {},
+        'portalSkin'           => 'bootstrap',
         'portalSkinBackground' =>
           '1280px-Cedar_Breaks_National_Monument_partially.jpg',
-        'mailUrl'                    => "http://auth.$domain/resetpwd",
         'localSessionStorage'        => 'Cache::FileCache',
         'localSessionStorageOptions' => {
             'namespace'          => 'lemonldap-ng-sessions',
@@ -193,6 +194,8 @@ sub zeroConf {
             'cache_root'         => "$cacheDir",
             'cache_depth'        => 3,
         },
+        "oidcServiceIgnoreScopeForClaims" => 1,
+        "webauthnDefaultTransports"       => "usb, nfc, ble, internal"
     };
 }
 
