@@ -8,7 +8,7 @@ use Lemonldap::NG::Common::FormEncode;
 use Lemonldap::NG::Common::UserAgent;
 use Lemonldap::NG::Portal::Main::Constants qw(PE_OK PE_ERROR PE_REDIRECT);
 
-our $VERSION = '2.0.6';
+our $VERSION = '2.19.0';
 
 extends 'Lemonldap::NG::Portal::Main::Auth';
 
@@ -84,7 +84,7 @@ sub extractFormInfo {
     my $nonce = time;
 
     # Build redirect_uri
-    my $callback_url = $self->conf->{portal};
+    my $callback_url = $req->portal;
 
     # Check return values
     my $error = $req->param("error");
@@ -261,19 +261,19 @@ sub setAuthSessionInfo {
           $req->data->{linkedInData}->{$_};
     }
 
-    PE_OK;
+    return PE_OK;
 }
 
 sub authenticate {
-    PE_OK;
+    return PE_OK;
 }
 
 sub authFinish {
-    PE_OK;
+    return PE_OK;
 }
 
 sub authLogout {
-    PE_OK;
+    return PE_OK;
 }
 
 sub authForce {
