@@ -9,7 +9,7 @@
 #
 package Lemonldap::NG::Portal::Main::Run;
 
-our $VERSION = '2.0.13';
+our $VERSION = '2.21.0';
 
 package Lemonldap::NG::Portal::Main;
 
@@ -17,7 +17,7 @@ use strict;
 use URI::Escape;
 use URI;
 use JSON;
-use Lemonldap::NG::Common::Util qw(getPSessionID);
+use Lemonldap::NG::Common::Util qw(getPSessionID getSameSite);
 
 has trOverCache => ( is => 'rw', default => sub { {} } );
 
@@ -36,11 +36,11 @@ sub authProcess { qw(extractFormInfo getUser authenticate) }
 
 sub sessionData {
     return qw(setAuthSessionInfo setSessionInfo), $_[0]->groupsAndMacros,
-      qw(setPersistentSessionInfo setLocalGroups store secondFactor);
+      qw(setPersistentSessionInfo setLocalGroups rememberBrowserCheck store secondFactor);
 }
 
 sub validSession {
-    qw(storeHistory buildCookie);
+    qw(storeHistory rememberBrowser buildCookie);
 }
 
 # RESPONSE HANDLER
@@ -54,6 +54,9 @@ sub handler {
 
     bless $req, 'Lemonldap::NG::Portal::Main::Request';
     $req->init( $self->conf );
+
+    # Set portal URL dynamically
+    $req->portal( HANDLER->tsv->{portal}->($req) );
     my $sp = 0;
 
     # Restore pdata
@@ -90,21 +93,41 @@ sub handler {
 
     # Save pdata
     if ( $sp or %{ $req->pdata } ) {
-        my %v = (
-            name   => $self->conf->{cookieName} . 'pdata',
-            secure => $self->conf->{securedCookie},
-            (
-                %{ $req->pdata }
-                ? ( value => uri_escape( JSON::to_json( $req->pdata ) ) )
-                : ( value => '', expires => 'Wed, 21 Oct 2015 00:00:00 GMT' )
-            ),
-            (
-                $self->conf->{pdataDomain}
-                ? ( domain => $self->conf->{pdataDomain}, )
-                : ()
-            ),
-        );
-        push @{ $res->[1] }, 'Set-Cookie', $self->cookie(%v);
+        my %pdata_options = ();
+        $pdata_options{domain} = $self->conf->{pdataDomain}
+          if $self->conf->{pdataDomain};
+
+        if ( %{ $req->pdata } ) {
+            push @{ $res->[1] }, 'Set-Cookie',
+              $self->cookie(
+                name   => $self->conf->{cookieName} . 'pdata',
+                secure => $self->conf->{securedCookie},
+                value  => uri_escape( JSON::to_json( $req->pdata ) ),
+                %pdata_options,
+              );
+
+        }
+        else {
+            push @{ $res->[1] }, 'Set-Cookie',
+              $self->cookie(
+                name    => $self->conf->{cookieName} . 'pdata',
+                secure  => $self->conf->{securedCookie},
+                value   => '',
+                expires => 'Wed, 21 Oct 2015 00:00:00 GMT',
+                %pdata_options,
+              );
+
+            # Avoid regressions with #3228
+            push @{ $res->[1] }, 'Set-Cookie',
+              $self->genDomainCookie(
+                $req,
+                name    => $self->conf->{cookieName} . 'pdata',
+                secure  => $self->conf->{securedCookie},
+                value   => '',
+                expires => 'Wed, 21 Oct 2015 00:00:00 GMT',
+                %pdata_options,
+              );
+        }
     }
     return $res;
 }
@@ -121,7 +144,20 @@ sub handler {
 
 sub authenticated {
     my ( $self, $req ) = @_;
-    return $self->sendJSONresponse( $req, { status => 1 } );
+
+    return $self->do(
+        $req,
+        [
+            'importHandlerData',
+            'controlUrl',
+            @{ $self->forAuthUser },
+            sub {
+                $req->response(
+                    $self->sendJSONresponse( $req, { status => 1 } ) );
+                return PE_SENDRESPONSE;
+            }
+        ]
+    );
 }
 
 sub pleaseAuth {
@@ -134,11 +170,11 @@ sub login {
     return $self->do(
         $req,
         [
-            'checkUnauthLogout', 'controlUrl',    # Fix 2342
-            @{ $self->beforeAuth },         $self->authProcess,
-            @{ $self->betweenAuthAndData }, $self->sessionData,
-            @{ $self->afterData },          $self->validSession,
-            @{ $self->endAuth }
+            'checkUnauthLogout', 'checkCancel',
+            'controlUrl',        @{ $self->beforeAuth },
+            $self->authProcess,  @{ $self->betweenAuthAndData },
+            $self->sessionData,  @{ $self->afterData },
+            $self->validSession, @{ $self->endAuth }
         ]
     );
 }
@@ -148,7 +184,7 @@ sub postLogin {
     return $self->do(
         $req,
         [
-            'checkUnauthLogout', 'restoreArgs',            # Fix 2342
+            'checkUnauthLogout', 'restoreArgs',                    # Fix 2342
             'controlUrl',        @{ $self->beforeAuth },
             $self->authProcess,  @{ $self->betweenAuthAndData },
             $self->sessionData,  @{ $self->afterData },
@@ -159,6 +195,7 @@ sub postLogin {
 
 sub authenticatedRequest {
     my ( $self, $req ) = @_;
+    $req->data->{alreadyAuthenticated} = 1;
     return $self->do(
         $req,
         [
@@ -180,35 +217,83 @@ sub postAuthenticatedRequest {
     );
 }
 
-sub refresh {
+sub processRefreshSession {
     my ( $self, $req ) = @_;
-    $req->mustRedirect(1);
+
     my %data = %{ $req->userData };
-    $self->userLogger->notice(
-        'Refresh request for ' . $data{ $self->conf->{whatToTrace} } );
     $req->user( $data{_user} || $data{ $self->conf->{whatToTrace} } );
     $req->id( $data{_session_id} );
+
+    $self->auditLog(
+        $req,
+        message => ( 'Refresh request for ' . $req->user ),
+        code    => "SESSION_REFRESH",
+        user    => $req->user,
+    );
+
     foreach ( keys %data ) {
-        delete $data{$_} unless ( /^_/ or /^(?:startTime)$/ );
+
+        # Variables that start with _ are kept accross refresh
+        if (/^_/) {
+
+            # But not OIDC tokens, which can be refreshed
+            delete $data{$_}
+              if (
+/^(_oidc_access_token|_oidc_refresh_token|_oidc_access_token_eol)$/
+              );
+        }
+
+        # Other variables should be refreshed
+        else {
+            # But not these two
+            if (/^(?:startTime|authenticationLevel)$/) {
+                next;
+            }
+            else {
+                delete $data{$_};
+            }
+        }
     }
     $data{_updateTime} = strftime( "%Y%m%d%H%M%S", localtime() );
     $self->logger->debug(
         "Set session $req->{id} _updateTime with $data{_updateTime}");
+
+    # Avoid interferences when refresh is run on multiple sessions
+    # in the same request
+    $req->sessionInfo( {} );
     $req->steps( [
             'getUser',
             @{ $self->betweenAuthAndData },
             'setSessionInfo',
-            $self->groupsAndMacros,
-            'setLocalGroups',
             sub {
                 $_[0]->sessionInfo->{$_} = $data{$_} foreach ( keys %data );
                 $_[0]->refresh(1);
                 return PE_OK;
             },
-            'store',
+            $self->groupsAndMacros,
+            'setLocalGroups',
         ]
     );
-    my $res = $req->error( $self->process($req) );
+    return $self->process($req);
+}
+
+sub refresh {
+    my ( $self, $req ) = @_;
+    $req->mustRedirect(1);
+
+    $self->userLogger->notice( 'Refresh request for '
+          . $req->userData->{ $self->conf->{whatToTrace} } );
+
+    my $res = $self->processRefreshSession($req);
+
+    # Run additional 'store' step We need to call process() here in case
+    # 'store' is hooked by aroundSub or afterSub
+    if ( $res == PE_OK ) {
+        $req->steps( ['store'] );
+        $res = $self->process($req);
+    }
+
+    $req->error($res);
     if ($res) {
         $req->info(
             $self->loadTemplate(
@@ -216,7 +301,7 @@ sub refresh {
                 'simpleInfo', params => { trspan => 'rightsReloadNeedsLogout' }
             )
         );
-        $req->urldc( $self->conf->{portal} );
+        $req->urldc( $req->portal );
         return $self->do( $req, [ sub { PE_INFO } ] );
     }
     return $self->do( $req, [ sub { PE_OK } ] );
@@ -230,73 +315,70 @@ sub logout {
             'importHandlerData',      'controlUrl',
             @{ $self->beforeLogout }, 'authLogout',
             'deleteSession'
-        ]
+        ],
+        1,    # No fail: logout should continue even if errors
     );
 }
 
 sub unauthLogout {
     my ( $self, $req ) = @_;
-    $self->userLogger->info('Unauthenticated logout request');
+    $self->_unauthLogout($req);
+    return $self->do( $req, [ 'controlUrl', sub { PE_LOGOUT_OK } ] );
+}
+
+sub _unauthLogout {
+    my ( $self, $req ) = @_;
+    $self->processHook( $req, 'unAuthLogout' );
+    $self->logger->debug('Unauthenticated logout request');
     $self->logger->debug('Cleaning pdata');
     $self->logger->debug("Removing $self->{conf}->{cookieName} cookie");
     $req->pdata( {} );
     $req->addCookie(
-        $self->cookie(
+        $self->genDomainCookie(
+            $req,
             name    => $self->conf->{cookieName},
-            domain  => $self->conf->{domain},
             secure  => $self->conf->{securedCookie},
             expires => 'Wed, 21 Oct 2015 00:00:00 GMT',
             value   => 0
         )
     );
-    return $self->do( $req, [ sub { PE_LOGOUT_OK } ] );
 }
 
 # RUNNING METHODS
 # ---------------
 
 sub do {
-    my ( $self, $req, $steps ) = @_;
+    my ( $self, $req, $steps, $nofail ) = @_;
     $req->steps($steps);
     $req->data->{activeTimer} = $self->conf->{activeTimer};
-    my $err = $req->error( $self->process($req) );
+    my $err = $req->error( $self->process( $req, nofail => $nofail ) );
 
     # Update status
-    if ( my $p = $self->HANDLER->tsv->{statusPipe} ) {
-        $p->print( ( $req->user ? $req->user : $req->address ) . ' => '
-              . $req->uri
-              . " $err\n" );
-    }
+    $self->HANDLER->publishStatus(
+        user       => ( $req->user ? $req->user : $req->address ),
+        url        => $req->{env}->{HTTP_HOST} . $req->{env}->{REQUEST_URI},
+        portalCode => $err,
+    );
 
     # Update history
-    if ( $err == PE_SENDRESPONSE ) {
-        return $req->response;
-    }
+    return $req->response if $err == PE_SENDRESPONSE;
 
     # Remove userData if authentication fails
-    if ( $err == PE_BADCREDENTIALS or $err == PE_BADOTP ) {
-        $req->userData( {} );
-    }
+    $req->userData( {} ) if ( $err == PE_BADCREDENTIALS or $err == PE_BADOTP );
 
     if ( !$self->conf->{noAjaxHook} and $req->wantJSON ) {
         $self->logger->debug('Processing to JSON response');
         if ( ( $err > 0 and !$req->id ) or $err eq PE_SESSIONNOTGRANTED ) {
             my $json = { result => 0, error => $err };
             if ( $req->wantErrorRender ) {
-                $json->{html} = $self->loadTemplate(
-                    $req,
-                    'errormsg',
-                    params => {
-                        AUTH_ERROR      => $err,
-                        AUTH_ERROR_TYPE => $req->error_type,
-                    }
-                );
+                $json->{html} = $self->loadTemplate( $req, 'errormsg',
+                    params => { $self->getErrorTplParams($req) } );
             }
             return $self->sendJSONresponse(
                 $req, $json,
                 code    => 401,
                 headers => [
-                    'WWW-Authenticate' => "SSO " . $self->conf->{portal},
+                    'WWW-Authenticate' => "SSO " . $req->portal,
                     "Content-Type"     => "application/json"
                 ],
             );
@@ -304,24 +386,26 @@ sub do {
         elsif ( $err > 0 and $err != PE_PASSWORD_OK and $err != PE_LOGOUT_OK ) {
             return $self->sendJSONresponse(
                 $req,
-                { result => 0, error => $err },
+                {
+                    result => 0,
+                    error  => $err,
+                    (
+                        $err == PE_NOTIFICATION && $req->id
+                        ? ( ciphered_id => $req->id )
+                        : ()
+                    )
+                },
                 code => 400
             );
         }
         else {
-            return $self->sendJSONresponse(
-                $req,
-                {
-                    result => 1,
-                    error  => $err,
-                    id     => $req->id,
-                    (
-                        $req->sessionInfo->{_httpSession}
-                        ? ( id_http => $req->sessionInfo->{_httpSession} )
-                        : ()
-                    )
-                }
-            );
+            my $res = { result => 1, error => $err };
+            unless ( $req->data->{alreadyAuthenticated} ) {
+                $res->{id}      = $req->id;
+                $res->{id_http} = $req->sessionInfo->{_httpSession}
+                  if $req->sessionInfo->{_httpSession};
+            }
+            return $self->sendJSONresponse( $req, $res );
         }
     }
     else {
@@ -355,6 +439,11 @@ sub do {
 
 sub getModule {
     my ( $self, $req, $type ) = @_;
+    if ( my $val =
+        $req->userData->{ { auth => '_auth', user => '_userDB' }->{$type} } )
+    {
+        return $val;
+    }
     if (
         my $mod = {
             auth     => '_authentication',
@@ -385,17 +474,32 @@ sub autoRedirect {
     my ( $self, $req ) = @_;
 
     # Set redirection URL if needed
-    $req->{urldc} ||= $self->conf->{portal}
-      if ( $req->mustRedirect and not( $req->info ) );
+    if ( $req->mustRedirect and not( $req->info ) ) {
+        if ( $req->error == PE_LOGOUT_OK ) {
+            if ( !$req->urldc or !$self->_isExternalUrl( $req, $req->urldc ) ) {
+                $req->urldc(
+                    $self->buildUrl( $req->portal, { "logout" => 1 } ) );
+            }
+        }
+        else {
+            $req->{urldc} ||= $req->portal;
+        }
+    }
 
     # Redirection should be made if urldc defined
     if ( $req->{urldc} ) {
-        $self->logger->debug("Building redirection to $req->{urldc}");
-        if ( $self->_jsRedirect->( $req, $req->sessionInfo ) ) {
-            $req->error(PE_REDIRECT);
-            $req->data->{redirectFormMethod} = "get";
+        if ( $req->{urldc} =~ /^\s*((?:java|vb)script|data):/ ) {
+            $self->auditLog(
+                $req,
+                message  => "Redirection to $req->{urldc} blocked",
+                code     => "UNAUTHORIZED_REDIRECT",
+                url      => $req->{urldc},
+                user     => $req->sessionInfo->{ $self->conf->{whatToTrace} } || $req->user,
+            );
+            delete $req->{urldc};
         }
         else {
+            $self->logger->debug("Building redirection to $req->{urldc}");
             if (    $req->{pdata}->{_url}
                 and $req->{pdata}->{_url} eq encode_base64( $req->{urldc}, '' )
               )
@@ -403,7 +507,20 @@ sub autoRedirect {
                 $self->logger->info("Force cleaning pdata");
                 delete $req->{pdata}->{_url};
             }
-            return [ 302, [ Location => $req->{urldc}, $req->spliceHdrs ], [] ];
+            if ( $self->_jsRedirect->( $req, $req->sessionInfo ) ) {
+                $req->error(PE_REDIRECT);
+                $req->data->{redirectFormMethod} = "get";
+            }
+            else {
+                return [
+                    302,
+                    [
+                        Location => URI->new( $req->{urldc} )->as_string,
+                        $req->spliceHdrs
+                    ],
+                    []
+                ];
+            }
         }
     }
     my ( $tpl, $prms ) = $self->display($req);
@@ -424,6 +541,11 @@ sub getApacheSession {
     }
 
     my $as = Lemonldap::NG::Common::Session->new( {
+
+            # The $args{hashStore} permits one to override the
+            # hashedSessionStore parameter. It is used when sessions are given
+            # by searchOn (like in Refresh plugin)
+            hashStore => $args{hashStore} // $self->conf->{hashedSessionStore},
             storageModule        => $self->conf->{globalStorage},
             storageModuleOptions => $self->conf->{globalStorageOptions},
             cacheModule          => $self->conf->{localSessionStorage},
@@ -451,7 +573,8 @@ sub getApacheSession {
         $self->logger->debug("Session $args{kind} $id not found");
         return;
     }
-    $self->logger->debug("Get session $id from Portal::Main::Run") if ($id);
+    $self->logger->debug("Get session $id from Portal::Main::Run")
+      if ($id);
     $self->logger->debug(
         "Check session validity  -> " . $self->conf->{timeoutActivity} . "s" )
       if ( $self->conf->{timeoutActivity} );
@@ -483,7 +606,9 @@ sub getApacheSession {
 sub getPersistentSession {
     my ( $self, $uid, $info ) = @_;
 
-    return unless ( defined $uid and !$self->conf->{disablePersistentStorage} );
+    return
+      unless ( defined $uid
+        and !$self->conf->{disablePersistentStorage} );
 
     # Compute persistent identifier
     my $pid = getPSessionID($uid);
@@ -583,7 +708,9 @@ sub updateSession {
             $self->logger->debug("Update sessionInfo $_");
             $self->_dump( $infos->{$_} );
             $req->{sessionInfo}->{$_} = $infos->{$_};
-            if ( $self->HANDLER->data->{_session_id} && $id eq $self->HANDLER->data->{_session_id} ) {
+            if (   $self->HANDLER->data->{_session_id}
+                && $id eq $self->HANDLER->data->{_session_id} )
+            {
                 $self->HANDLER->data->{$_} = $infos->{$_};
             }
         }
@@ -598,6 +725,9 @@ sub updateSession {
                 $self->logger->error( $apacheSession->error );
             }
         }
+
+        # remove the corresponding session from handler cache
+        HANDLER->publishEvent( $req, { action => 'unlog', id => $id } );
     }
 }
 
@@ -628,25 +758,27 @@ sub _deleteSession {
 
         # Create an obsolete cookie to remove it
         $req->addCookie(
-            $self->cookie(
+            $self->genDomainCookie(
+                $req,
                 name    => $self->conf->{cookieName} . 'http',
                 value   => 0,
-                domain  => $self->conf->{domain},
                 secure  => 0,
                 expires => 'Wed, 21 Oct 2015 00:00:00 GMT'
             )
         ) unless ($preserveCookie);
     }
 
-    HANDLER->localUnlog( $req, $session->id );
+    # Publishing an "unlog" event will automatically remove the corresponding
+    # session from handler cache. See publishEvent into Handler::Main::Run
+    HANDLER->publishEvent( $req, { action => 'unlog', id => $session->id } );
     $session->remove;
 
     # Create an obsolete cookie to remove it
     $req->addCookie(
-        $self->cookie(
+        $self->genDomainCookie(
+            $req,
             name    => $self->conf->{cookieName},
             value   => 0,
-            domain  => $self->conf->{domain},
             secure  => $self->conf->{securedCookie},
             expires => 'Wed, 21 Oct 2015 00:00:00 GMT'
         )
@@ -655,9 +787,17 @@ sub _deleteSession {
     # Log
     my $user = $req->{sessionInfo}->{ $self->conf->{whatToTrace} };
     my $mod  = $req->{sessionInfo}->{_auth};
-    $self->userLogger->notice(
+    if ($user) {
+        $self->auditLog(
+            $req,
+            message => (
 "User $user has been disconnected from $mod ($req->{sessionInfo}->{ipAddr})"
-    ) if $user;
+            ),
+            code     => "LOGOUT",
+            user     => $user,
+            auth     => $mod,
+        );
+    }
 
     return $session->error ? 0 : 1;
 }
@@ -686,6 +826,16 @@ sub autoPost {
 
     # Get URL and Form fields
     $req->{urldc} = $req->postUrl;
+    if ( $req->{urldc} =~ /^\s*((?:java|vb)script|data):/ ) {
+        $self->auditLog(
+            $req,
+            message  => "Redirection to $req->{urldc} blocked",
+            code     => "UNAUTHORIZED_REDIRECT",
+            url      => $req->{urldc},
+            user     => $req->sessionInfo->{ $self->conf->{whatToTrace} } || $req->user,
+        );
+        return PE_BADURL;
+    }
     my $formFields = $req->postFields;
 
     $self->clearHiddenFormValue($req);
@@ -700,6 +850,13 @@ sub autoPost {
     }
 
     $req->data->{redirectFormMethod} = "post";
+
+    # Case 2F wait
+    if ( $req->data->{sfWait} ) {
+        return PE_2FWAIT;
+    }
+
+    # Default case -> redirect
     return PE_REDIRECT;
 }
 
@@ -793,22 +950,47 @@ sub info {
 
 sub fullUrl {
     my ( $self, $req ) = @_;
-    my $pHost = $self->conf->{portal};
+    my $pHost = $req->portal;
     $pHost =~ s#^(https?://[^/]+)(?:/.*)?$#$1#;
     return $pHost . $req->env->{REQUEST_URI};
 }
 
+# Generates a cookie header which can depend on the request
+# If no domain was explicitely specified, use the default SSO domain
+sub genDomainCookie {
+    my ( $self, $req, %h ) = @_;
+
+    $h{domain} ||= $self->getCookieDomain($req);
+    return $self->cookie(%h);
+}
+
+# DEPRECATED, #3228
+sub genCookie {
+    my ( $self, $req, %h ) = @_;
+    return $self->cookie(%h);
+}
+
+# Generate a cookie header
+# If no domain was explicitely specified, only the portal will see
+# this cookie
 sub cookie {
     my ( $self, %h ) = @_;
-    my @res;
-    $res[0] = "$h{name}" or die("name required");
-    $res[0] .= "=$h{value}";
     $h{path} ||= '/';
     $h{HttpOnly} //= $self->conf->{httpOnly};
     $h{max_age}  //= $self->conf->{cookieExpiration}
       if ( $self->conf->{cookieExpiration} );
     $h{SameSite} ||= $self->cookieSameSite;
 
+    return $self->assemble_cookie(%h);
+}
+
+sub assemble_cookie {
+    my ( $self, %h ) = @_;
+
+    my @res;
+    $res[0] = "$h{name}" or die("name required");
+    $res[0] .= "=$h{value}";
+
     foreach (qw(domain path expires max_age HttpOnly SameSite)) {
         my $f = $_;
         $f =~ s/_/-/g;
@@ -829,26 +1011,24 @@ sub _dump {
     return;
 }
 
-sub sendHtml {
-    my ( $self, $req, $template, %args ) = @_;
+sub getSkinTplDir {
+    my ( $self, $skin ) = @_;
 
-    my $templateDir = $self->conf->{templateDir} . '/' . $self->getSkin($req);
-    $self->templateDir( [ $templateDir, @{ $self->templateDir } ] );
+    my $base = $self->conf->{skinTemplateDir} || $self->conf->{templateDir};
+    return ( $base . '/' . $skin );
+}
 
-    # Check template
-    $args{templateDir} = $templateDir;
-    my $tmpl = $args{templateDir} . "/$template.tpl";
-    unless ( -f $tmpl ) {
-        $self->logger->debug("Template $tmpl not found");
-        $args{templateDir} = $self->conf->{templateDir} . '/bootstrap';
-        $tmpl = $args{templateDir} . "/$template.tpl";
-        $self->logger->debug("-> Trying to load $tmpl");
-    }
+# Warning: this function returns a JSON string
+sub getTrOver {
+    my ( $self, $req, $templateDir ) = @_;
 
-    # Override messages
-    my $trOverMessages = JSON::from_json( $self->trOver );
+    $templateDir //= $self->getSkinTplDir( $self->getSkin($req) );
 
     unless ( $self->trOverCache->{$templateDir} ) {
+
+        # Override messages
+        my $trOverMessages = JSON::from_json( $self->trOver );
+
         opendir( DIR, $templateDir );
         my @langfiles = grep( /\.json$/, readdir(DIR) );
         close(DIR);
@@ -857,8 +1037,21 @@ sub sendHtml {
             my ($lang) = ( $file =~ /^(\w+)\.json/ );
             $self->logger->debug("Use $file to override messages");
             if ( open my $json, "<", $templateDir . "/" . $file ) {
-                local $/ = undef;
-                $trOverMessages->{$lang} = JSON::from_json(<$json>);
+                my $trdata;
+                eval {
+                    local $/ = undef;
+                    $trdata = JSON::from_json(<$json>);
+                };
+                if ($@) {
+                    $self->logger->warn("Ignoring $file because of error: $@");
+                }
+                if ( ref($trdata) eq "HASH" ) {
+                    for my $msg ( keys %$trdata ) {
+
+                        # lemonldap-ng.ini has priority
+                        $trOverMessages->{$lang}->{$msg} //= $trdata->{$msg};
+                    }
+                }
             }
             else {
                 $self->logger->error("Unable to read $file");
@@ -867,7 +1060,26 @@ sub sendHtml {
 
         $self->trOverCache->{$templateDir} = JSON::to_json($trOverMessages);
     }
-    $args{params}->{TROVER} = $self->trOverCache->{$templateDir};
+
+    return $self->trOverCache->{$templateDir};
+}
+
+sub sendHtml {
+    my ( $self, $req, $template, %args ) = @_;
+    if ( $self->conf->{logParams} ) {
+        $self->_dump($req);
+        $self->_dump( \%args );
+    }
+
+    my $skin_template_dir = $self->getSkinTplDir( $self->getSkin($req) );
+
+    # Look for templates in skin subdirectories instead of templateDir
+    $args{templateDir} =
+      [ $skin_template_dir, $self->conf->{templateDir} . "/bootstrap" ];
+
+    $args{params}->{TROVER} = $self->getTrOver( $req, $skin_template_dir );
+
+    $self->processHook( $req, 'sendHtml', \$template, \%args );
 
     my $res = $self->SUPER::sendHtml( $req, $template, %args );
     push @{ $res->[1] },
@@ -879,22 +1091,34 @@ sub sendHtml {
 
     $self->setCorsHeaderFromConfig($res);
 
+    if (    $self->conf->{strictTransportSecurityMax_Age}
+        and $req->portal =~ /^https:/ )
+    {
+        push @{ $res->[1] },
+          'Strict-Transport-Security' =>
+          "max-age=$self->{conf}->{strictTransportSecurityMax_Age}";
+        $self->logger->debug(
+"Set Strict-Transport-Security with: $self->{conf}->{strictTransportSecurityMax_Age}"
+        );
+    }
+
     # Set authorized URL for POST
     my $csp = $self->csp . "form-action " . $self->conf->{cspFormAction};
     if ( my $url = $req->urldc ) {
-        $self->logger->debug("Required urldc : $url");
-        $url =~ URIRE;
-        $url = $2 . '://' . $3 . ( $4 ? ":$4" : '' );
-        $self->logger->debug("Set CSP form-action with urldc : $url");
-        $csp .= " $url";
+        $self->logger->debug("Required urldc: $url");
+        my $host = $self->cspGetHost($url);
+        if ($host) {
+            $self->logger->debug("Set CSP form-action with urldc: $url");
+            $csp .= " $host";
+        }
     }
     my $url = $args{params}->{URL};
     if ( defined $url ) {
-        $self->logger->debug("Required Params URL : $url");
-        if ( $url =~ URIRE ) {
-            $url = $2 . '://' . $3 . ( $4 ? ":$4" : '' );
-            $self->logger->debug("Set CSP form-action with Params URL : $url");
-            $csp .= " $url";
+        $self->logger->debug("Required Params URL: $url");
+        my $host = $self->cspGetHost($url);
+        if ($host) {
+            $self->logger->debug("Set CSP form-action with Params URL: $url");
+            $csp .= " $host";
         }
     }
     if ( defined $req->data->{cspFormAction}
@@ -914,7 +1138,6 @@ sub sendHtml {
     {
         $self->logger->debug(
             "Add SAML Discovery Protocol URL in CSP form-action");
-
         $csp .= " " . $self->conf->{samlDiscoveryProtocolURL};
     }
     $csp .= ';';
@@ -938,7 +1161,7 @@ sub sendHtml {
     # created in the code, and remove this
     my @url;
     if ( $req->info ) {
-        @url = map { s#https?://([^/]+).*#$1#; $_ }
+        @url = map { s#https?://([^/]+).*#$1#r }
           ( $req->info =~ /<iframe.*?src="(.*?)"/sg );
     }
 
@@ -951,11 +1174,36 @@ sub sendHtml {
     }
 
     # Set CSP header
-    push @{ $res->[1] }, 'Content-Security-Policy' => $csp;
-    $self->logger->debug("Apply following CSP : $csp");
+    push @{ $res->[1] }, 'Content-Security-Policy' => $csp
+      unless $req->data->{dropCsp};
+    $self->logger->debug("Apply following CSP: $csp");
     return $res;
 }
 
+sub imgok {
+    my ( $self, $req, ) = @_;
+    return $self->sendImage( $req, 'icons/ok.png' );
+}
+
+sub imgnok {
+    my ( $self, $req, ) = @_;
+    return $self->sendImage( $req, 'icons/warning.png' );
+}
+
+sub sendImage {
+    my ( $self, $req, $img ) = @_;
+    my $u = URI->new_abs( $self->staticPrefix . "/common/$img", $req->portal )
+      ->as_string;
+
+    return [
+        302,
+        [
+            'Location' => $u,
+        ],
+        [],
+    ];
+}
+
 sub sendCss {
     my ( $self, $req ) = @_;
     my $s = '/* LL::NG Portal CSS */';
@@ -998,7 +1246,7 @@ sub lmError {
     my %templateParams = (
         MAIN_LOGO  => $self->conf->{portalMainLogo},
         LANGS      => $self->conf->{showLanguages},
-        LOGOUT_URL => $self->conf->{portal} . "?logout=1",
+        LOGOUT_URL => $self->buildUrl( $req->portal, { logout => 1 } ),
         URL        => $req->{urldc},
     );
 
@@ -1023,7 +1271,7 @@ sub tplParams {
     my ( $self, $req ) = @_;
     my %templateParams;
 
-    my $portalPath = $self->conf->{portal};
+    my $portalPath = $req->portal;
     $portalPath =~ s#^https?://[^/]+/?#/#;
     $portalPath =~ s#[^/]+\.fcgi$##;
 
@@ -1037,16 +1285,45 @@ sub tplParams {
     }
 
     return (
-        SKIN       => $self->getSkin($req),
-        PORTAL_URL => $self->conf->{portal},
-        SKIN_PATH  => $portalPath . "skins",
-        SKIN_BG    => $self->conf->{portalSkinBackground},
-        CUSTOM_CSS => $self->conf->{portalCustomCss},
-        ( $self->customParameters ? ( %{ $self->customParameters } ) : () ),
+        CACHE_TAG    => $self->cacheTag,
+        PORTAL_URL   => $req->portal,
+        PORTAL_BASE  => $portalPath,
+        MAIN_LOGO    => $self->conf->{portalMainLogo},
+        LANGS        => $self->conf->{showLanguages},
+        SCROLL_TOP   => $self->conf->{scrollTop},
+        FLOATING_CAT => $self->conf->{floatingCategoryName},
+        SKIN         => $self->getSkin($req),
+        SKIN_PATH    => $portalPath . "skins",
+        SAMESITE     => getSameSite( $self->conf ),
+        COOKIESECURE => ( $self->conf->{securedCookie} ? 1 : 0 ),
+        SKIN_BG      => $self->conf->{portalSkinBackground},
+        FAVICON      => $self->conf->{portalFavicon} || 'common/favicon.ico',
+        LANGUAGE     => $self->getLanguage($req),
+        CUSTOM_CSS   => $self->conf->{portalCustomCss},
+        CUSTOM_JS    => $self->conf->{portalCustomJs},
+        (
+            $self->customParameters
+            ? ( %{ $self->customParameters } )
+            : ()
+        ),
         %templateParams
     );
 }
 
+sub getErrorTplParams {
+    my ( $self, $req ) = @_;
+
+    my $err = $req->error;
+
+    return (
+        AUTH_ERROR               => $err,
+        ( 'AUTH_ERROR_' . $err ) => 1,
+        AUTH_ERROR_TYPE          => $req->error_type,
+        AUTH_ERROR_ROLE          => $req->error_role,
+        LOCKTIME                 => $req->lockTime(),
+    );
+}
+
 sub registerLogin {
     my ( $self, $req ) = @_;
     return
@@ -1083,12 +1360,13 @@ sub registerLogin {
     }
 
     my $history = $req->sessionInfo->{_loginHistory} ||= {};
-    my $type = ( $req->authResult > 0 ? 'failed' : 'success' ) . 'Login';
+    my $type    = ( $req->authResult > 0 ? 'failed' : 'success' ) . 'Login';
     $history->{$type} ||= [];
     $self->logger->debug("Current login saved into $type");
 
-    # Gather current login's parameters
-    my $login = $self->_sumUpSession( $req->{sessionInfo}, 1 );
+    # Gather current login's parameters and force current time
+    my $login =
+      $self->_sumUpSession( { %{ $req->sessionInfo }, _utime => time }, 1 );
     $login->{error} = $self->error( $req->authResult )
       if ( $req->authResult );
 
@@ -1108,6 +1386,34 @@ sub registerLogin {
     PE_OK;
 }
 
+# Register access to protected app
+# @param Lemonldap::NG::Portal::Main::Request $req The request
+# @param string $uid The connected user ident
+# @param string $app The application client/RP namespaced with protocol
+sub registerProtectedAppAccess {
+    my ( $self, $req, $uid, $app ) = @_;
+
+    return unless $self->conf->{appAccessHistoryEnabled};
+
+    $self->logger->debug("Registering $app access for user $uid");
+
+    my $persistentSession = $self->getPersistentSession($uid);
+    my $entry = $persistentSession->data->{_appHistory}->{$app} || {};
+
+    $entry->{access_time} = time();
+    $entry->{access_count}++;
+
+    $self->updatePersistentSession(
+        $req,
+        {
+            _appHistory => {
+                %{ $persistentSession->data->{_appHistory} }, $app => $entry
+            }
+        },
+        $uid
+    );
+}
+
 # put main session data into a hash ref
 # @param hashref $session The session to sum up
 # @return hashref
@@ -1118,8 +1424,15 @@ sub _sumUpSession {
       ? {}
       : { user => $session->{ $self->conf->{whatToTrace} } };
     $res->{$_} = $session->{$_}
-      foreach ( "_utime", "ipAddr",
-        keys %{ $self->conf->{sessionDataToRemember} } );
+      foreach (
+        "_utime", "ipAddr",
+        map {
+            # Modifying key to remove ordering prefix
+            $_ =~ s/(\d+_)?//;
+            $_;
+        } keys %{ $self->conf->{sessionDataToRemember} },
+        keys %{ $self->pluginSessionDataToRemember }
+      );
     return $res;
 }
 
@@ -1135,6 +1448,7 @@ sub corsPreflight {
 
 sub sendJSONresponse {
     my ( $self, $req, $j, %args ) = @_;
+    $j->{token} = $req->token if $req->token;
     my $res = Lemonldap::NG::Common::PSGI::sendJSONresponse(@_);
 
     # Handle caching
@@ -1151,7 +1465,7 @@ sub sendJSONresponse {
     # If this is a cross-domain request from the portal itself
     # (Ajax SSL to a different VHost)
     # we allow CORS
-    if ( $req->origin and index( $self->conf->{portal}, $req->origin ) == 0 ) {
+    if ( $self->_checkSelfCors($req) ) {
         $self->logger->debug('AJAX request from portal, allowing CORS');
         push @{ $res->[1] },
           "Access-Control-Allow-Origin"      => $req->origin,
@@ -1165,6 +1479,21 @@ sub sendJSONresponse {
     return $res;
 }
 
+sub _checkSelfCors {
+    my ( $self, $req ) = @_;
+
+    if ( $req->origin ) {
+        my $origin = URI->new( $req->origin );
+        my $portal = URI->new( $req->portal );
+
+        return (  $origin->scheme
+              and $portal->scheme eq $origin->scheme
+              and $origin->host_port
+              and $origin->host_port eq $portal->host_port );
+    }
+    return;
+}
+
 sub sendRawHtml {
     my ($self) = $_[0];
     my $res = Lemonldap::NG::Common::PSGI::sendRawHtml(@_);
@@ -1180,7 +1509,7 @@ sub setCorsHeaderFromConfig {
     if ( $self->conf->{corsEnabled} ) {
         my @cors = split /;/, $self->cors;
         push @{ $response->[1] }, @cors;
-        $self->logger->debug('Apply following CORS policy :');
+        $self->logger->debug('Apply following CORS policy:');
         $self->logger->debug(" $_") for @cors;
     }
 }
@@ -1188,11 +1517,13 @@ sub setCorsHeaderFromConfig {
 # Temlate loader
 sub loadTemplate {
     my ( $self, $req, $name, %prm ) = @_;
+
+    $self->processHook( $req, 'sendHtml', \$name, \%prm );
     $name .= '.tpl';
     my $tpl = HTML::Template->new(
         filename => $name,
         path     => [
-            $self->conf->{templateDir} . '/' . $self->getSkin($req),
+            $self->getSkinTplDir( $self->getSkin($req) ),
             $self->conf->{templateDir} . '/bootstrap/',
             $self->conf->{templateDir} . '/common/'
         ],
@@ -1203,6 +1534,9 @@ sub loadTemplate {
         global_vars            => 0,
         ( $prm{filter} ? ( filter => $prm{filter} ) : () ),
     );
+    if ( $self->can('tplParams') ) {
+        $tpl->param( $self->tplParams($req) );
+    }
     if ( $prm{params} ) {
         $tpl->param( %{ $prm{params} } );
     }
@@ -1213,12 +1547,92 @@ sub loadTemplate {
 # Content-Security-Polity header
 sub cspGetHost {
     my ( $self, $url ) = @_;
-    my $uri = $url // "";
-    unless ( $uri->isa("URI") ) {
+    return unless $url;
+    my $uri = $url;
+    unless ( blessed($uri) && $uri->isa("URI") ) {
         $uri = URI->new($uri);
     }
-    return (
-        $uri->scheme . "://" . ( $uri->_port ? $uri->host_port : $uri->host ) );
+    my $scheme = $uri->scheme || "";
+    if ( $scheme =~ /^https?/ ) {
+        return (
+            $scheme . "://"
+              . (
+                ( $uri->port == $uri->default_port )
+                ? $uri->host
+                : $uri->host_port
+              )
+        );
+    }
+    elsif ($scheme) {
+        return ( $scheme . ":" );
+    }
+    return;
+}
+
+sub relativeUrl {
+    my $self = shift;
+    my $req  = shift;
+
+    my $uri = $self->_buildUrlObj( $req->portal, @_ );
+    return $uri->path_query;
+}
+
+sub buildUrl {
+    my $self = shift;
+    return $self->portal unless @_;
+
+    return $self->_buildUrlObj(@_)->as_string;
+}
+
+sub _buildUrlObj {
+    my $self = shift;
+    return URI->new( $self->portal ) unless @_;
+
+    # URL base is $self->portal unless first arg is an URL
+    my $uri =
+      URI->new( ( $_[0] =~ m#^https?://# ) ? shift(@_) : $self->portal );
+    my @pathSg = grep { $_ ne '' } $uri->path_segments;
+    while (@_) {
+        my $s = shift;
+        if ( ref $s ) {
+            $uri->query_form($s);
+            if (@_) {
+                require Carp;
+                Carp::confess('Query must be the last arg of buildUrl');
+            }
+        }
+        else {
+            push @pathSg, $s;
+        }
+    }
+    $uri->path_segments(@pathSg);
+    return $uri;
+}
+
+sub rememberBrowser {
+    my ( $self, $req ) = @_;
+
+    if (    $self->getService('trustedBrowser')
+        and $self->getService('trustedBrowser')->can("newDevice") )
+    {
+        return $self->getService('trustedBrowser')->newDevice($req);
+    }
+    else {
+        return PE_OK;
+    }
+}
+
+sub rememberBrowserCheck {
+    my ( $self, $req ) = @_;
+
+    if (    $self->getService('trustedBrowser')
+        and $self->getService('trustedBrowser')->can("check") )
+    {
+        return $self->getService('trustedBrowser')->check($req);
+    }
+    else {
+        return PE_OK;
+    }
 }
 
 1;
